@@ -17,15 +17,21 @@ use nb::block;
 use stm32f1xx_hal::can::Can;
 use stm32f1xx_hal::i2c::{BlockingI2c, Mode};
 use stm32f1xx_hal::pac::{CorePeripherals, Peripherals, ITM};
+use stm32f1xx_hal::flash::{SectorSize, FlashSize};
+use stm32f1xx_hal::delay::Delay;
 use stm32f1xx_hal::prelude::*;
 
 use ross_eeprom::{RossEeprom, RossDeviceInfo};
+use ross_protocol::ross_packet::RossPacket;
 use ross_protocol::ross_convert_packet::RossConvertPacket;
 use ross_protocol::ross_event::ross_bootloader_event::RossBootloaderHelloEvent;
-use ross_protocol::ross_event::ross_programmer_event::RossProgrammerHelloEvent;
+use ross_protocol::ross_event::ross_programmer_event::*;
+use ross_protocol::ross_event::ross_general_event::*;
 use ross_protocol::ross_interface::ross_can::{RossCan, RossCanError};
 
-const PROGRAM_ADDRESS: u32 = 0x0800_8000;
+const PROGRAM_ADDRESS: u32 = 0x0801_0000;
+const FLASH_OFFSET: u32 = 0x0001_0000;
+const FLASH_SECTOR_SIZE: usize = 1024;
 
 const EEPROM_BITRATE: u32 = 400_000;
 
@@ -92,6 +98,7 @@ fn main() -> ! {
 
     let mut afio = dp.AFIO.constrain(&mut rcc.apb2);
     let mut gpiob = dp.GPIOB.split(&mut rcc.apb2);
+    let mut delay = Delay::new(cp.SYST, clocks);
 
     let mut eeprom = {
         let i2c1 = {
@@ -118,7 +125,6 @@ fn main() -> ! {
     };
 
     let device_info = eeprom.read_device_info().unwrap();
-
     debug!("Loaded device information from EEPROM ({:?}).", device_info);
 
     let mut can = {
@@ -149,8 +155,82 @@ fn main() -> ! {
 
     allocate_heap();
 
-    let programmer_hello_event = wait_for_programmer_hello_event(&mut can);
-    transmit_bootloader_hello_event(&mut can, &device_info, &programmer_hello_event);
+    let mut flash_writer = flash.writer(SectorSize::Sz1K, FlashSize::Sz128K);
+
+    flash_writer.change_verification(true);
+
+    let mut firmware_size: u32 = 0;
+    let mut new_device_info = None;
+
+    let mut buf = [0x00; FLASH_SECTOR_SIZE];
+    let mut buf_offset = 0;
+    let mut flash_offset = FLASH_OFFSET;
+
+    loop {
+        let packet = wait_for_packet(&mut can);
+
+        if let Ok(event) = RossProgrammerHelloEvent::try_from_packet(&packet) {    
+            debug!("Received `programmer_hello_event` ({:?}).", event);
+
+            transmit_bootloader_hello_event(&mut can, &device_info, &event);
+        } else if let Ok(event) = RossProgrammerStartUploadEvent::try_from_packet(&packet) {
+            debug!("Received `programmer_start_upload_event` ({:?}).", event);
+
+            firmware_size = event.firmware_size;
+            new_device_info = Some(RossDeviceInfo {
+                device_address: device_info.device_address,
+                firmware_version: event.new_firmware_version,
+                peripheral_info_address: device_info.peripheral_info_address,
+                program_info_address: device_info.program_info_address,
+            });
+
+            transmit_ack_event(&mut can, &device_info, event.programmer_address);
+        } else if let Ok(event) = RossDataEvent::try_from_packet(&packet) {
+            if let Some(ref device_info) = new_device_info {
+                transmit_ack_event(&mut can, &device_info, event.transmitter_address);
+                
+                for i in 0..event.data_len {
+                    buf[buf_offset + i as usize] = event.data[i as usize];
+                }
+
+                buf_offset += event.data_len as usize;
+
+                if buf_offset >= buf.len() {
+                    if buf_offset != buf.len() {
+                        debug!("Data packet length is not a divisor of {}.", buf.len());
+                        break;
+                    }
+
+                    if let Err(err) = flash_writer.erase(flash_offset as u32, buf.len()) {
+                        debug!("Failed to erase FLASH at offset {:#010x} with error ({:?}).", flash_offset, err);
+                        break;
+                    }
+
+                    if let Err(err) = flash_writer.write(flash_offset as u32, &buf[..]) {
+                        debug!("Failed to write to FLASH at offset {:#010x} with error ({:?}).", flash_offset, err);
+                        break;
+                    }
+
+                    flash_offset += buf.len() as u32;
+                    buf_offset = 0;
+
+                    if flash_offset == FLASH_OFFSET + firmware_size {
+                        debug!("Writing new device info to EEPROM ({:?}).", new_device_info);
+                        eeprom.write_device_info(&device_info, &mut delay);
+                        firmware_size = 0;
+                        new_device_info = None;
+
+                        debug!("Booting firmware.");
+                        boot();
+                    }
+                }
+            } else {
+                debug!("Unexpected `data_event` before `programmer_start_upload_event`.");
+            }
+        } else {
+            debug!("Received unexpected event ({:?}).", packet);
+        }
+    }
 
     loop {
         nop();
@@ -163,39 +243,20 @@ fn boot() -> ! {
     }
 }
 
-fn calc_can_btr(clock_rate: u32) -> u32 {
-    let brp = clock_rate / CAN_BITRATE / (CAN_TSEG1 + CAN_TSEG2);
-
-    (brp - 1) | ((CAN_TSEG1 - 1) << 16) | ((CAN_TSEG2 - 1) << 20) | ((CAN_SJW - 1) << 24)
-}
-
-fn allocate_heap() {
-    let start = cortex_m_rt::heap_start() as usize;
-    unsafe { ALLOCATOR.init(start, HEAP_SIZE) }
-}
-
-fn wait_for_programmer_hello_event<I: Instance>(can: &mut RossCan<I>) -> RossProgrammerHelloEvent {
+fn wait_for_packet<I: Instance>(can: &mut RossCan<I>) -> RossPacket {
     loop {
         let packet = match can.try_get_packet() {
-            Ok(packet) => packet,
+            Ok(packet) => return packet,
             Err(err) => {
                 if let RossCanError::NoPacketReceived = err {
                     continue;
                 }
 
                 debug!("Failed to get packet ({:?}).", err);
-                continue;
             },
         };
-
-        if let Ok(event) = RossProgrammerHelloEvent::try_from_packet(&packet) {    
-            debug!("Received 'programmer_hello_event' ({:?}).", event);
-            return event;
-        } else {
-            debug!("Received unexpected event.");
-        }
     }
-}
+} 
 
 fn transmit_bootloader_hello_event<I: Instance>(
     can: &mut RossCan<I>,
@@ -215,6 +276,30 @@ fn transmit_bootloader_hello_event<I: Instance>(
     }
 }
 
+fn transmit_ack_event<I: Instance>(can: &mut RossCan<I>, device_info: &RossDeviceInfo, transmitter_address: u16) {
+    let event = RossAckEvent {
+        device_address: device_info.device_address,
+        transmitter_address,
+    };
+
+    if let Err(err) = can.try_send_packet(&event.to_packet()) {
+        debug!("Failed to send `ack_event` ({:?}).", err);
+    } else {
+        debug!("Sent `ack_event` ({:?}).", event);
+    }
+}
+
+fn calc_can_btr(clock_rate: u32) -> u32 {
+    let brp = clock_rate / CAN_BITRATE / (CAN_TSEG1 + CAN_TSEG2);
+
+    (brp - 1) | ((CAN_TSEG1 - 1) << 16) | ((CAN_TSEG2 - 1) << 20) | ((CAN_SJW - 1) << 24)
+}
+
+fn allocate_heap() {
+    let start = cortex_m_rt::heap_start() as usize;
+    unsafe { ALLOCATOR.init(start, HEAP_SIZE) }
+}
+
 #[alloc_error_handler]
 fn oom(_: Layout) -> ! {
     loop {}
